import setuptools

setuptools.setup(
    name="crash-vm",
    version="0.0.1",
    description="Simple virtual machine implemented in pure python",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=["crash_vm"],
    python_requires='>=3.7',
)
